/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.openstack.nova.ec2.network;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.compute.ComputeServices;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.network.Direction;
import org.dasein.cloud.network.Firewall;
import org.dasein.cloud.network.FirewallRule;
import org.dasein.cloud.network.FirewallSupport;
import org.dasein.cloud.network.Protocol;
import org.dasein.cloud.openstack.nova.ec2.NovaEC2;
import org.dasein.cloud.openstack.nova.ec2.NovaException;
import org.dasein.cloud.openstack.nova.ec2.NovaMethod;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;

public class SecurityGroup implements FirewallSupport {
	static private final Logger logger = NovaEC2.getLogger(SecurityGroup.class, "std");
	
	private NovaEC2 provider = null;
	
	SecurityGroup(NovaEC2 provider) {
		this.provider = provider;
	}
	
	@Override
	public @Nonnull String authorize(@Nonnull String securityGroupId, @Nonnull String cidr, @Nonnull Protocol protocol, int startPort, int endPort) throws CloudException, InternalException {
        Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.AUTHORIZE_SECURITY_GROUP_INGRESS);
        NovaMethod method;
        NodeList blocks;
        Document doc;

        if( cidr.indexOf('/') == -1 ) {
            cidr = cidr + "/32";
        }
        parameters.put("GroupName", securityGroupId);
        parameters.put("IpProtocol", protocol.name().toLowerCase());
        parameters.put("FromPort", String.valueOf(startPort));
        parameters.put("ToPort", endPort == -1 ? String.valueOf(startPort) : String.valueOf(endPort));
        parameters.put("CidrIp", cidr);
        /*
        else {
            parameters.put("GroupId", firewallId);
            parameters.put("IpPermissions.1.IpProtocol", protocol.name().toLowerCase());
            parameters.put("IpPermissions.1.FromPort", String.valueOf(beginPort));
            parameters.put("IpPermissions.1.ToPort", endPort == -1 ? String.valueOf(beginPort) : String.valueOf(endPort));
            parameters.put("IpPermissions.1.IpRanges.1.CidrIp", cidr);
        }
        */
        method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
            doc = method.invoke();
        }
        catch( NovaException e ) {
            String code = e.getCode();

            if( code != null && code.equals("InvalidPermission.Duplicate") ) {
                return FirewallRule.getRuleId(securityGroupId, cidr, Direction.INGRESS, protocol, startPort, endPort);
            }
            logger.error(e.getSummary());
            throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("return");
        if( blocks.getLength() > 0 ) {
            if( !blocks.item(0).getFirstChild().getNodeValue().equalsIgnoreCase("true") ) {
                throw new CloudException("Failed to authorize security group rule without explanation.");
            }
        }
        return FirewallRule.getRuleId(securityGroupId, cidr, Direction.INGRESS, protocol, startPort, endPort);
    }

    @Override
	public @Nonnull String create(@Nonnull String name, @Nonnull String description) throws InternalException, CloudException {
		Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.CREATE_SECURITY_GROUP);
		NovaMethod method;
        NodeList blocks;
		Document doc;

		name = getUniqueName(name);
		parameters.put("GroupName", name);
		parameters.put("GroupDescription", description);
		method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
        	doc = method.invoke();
        }
        catch( NovaException e ) {
        	logger.error(e.getSummary());
        	throw new CloudException(e);
        }
        return name;
        /*
        }
        else {
            blocks = doc.getElementsByTagName("groupId");
            if( blocks.getLength() > 0 ) {
                return blocks.item(0).getFirstChild().getNodeValue().trim();
            }
            throw new CloudException("Failed to create security group without explanation.");
        }
        */
	}
 	
   @Override
    public @Nonnull String createInVLAN(@Nonnull String name, @Nonnull String description, @Nonnull String providerVlanId) throws InternalException, CloudException {
        Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.CREATE_SECURITY_GROUP);
        NovaMethod method;
        NodeList blocks;
        Document doc;

        name = getUniqueName(name);
        parameters.put("GroupName", name);
        parameters.put("GroupDescription", description);
        parameters.put("VpcId", providerVlanId);
        method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
            doc = method.invoke();
        }
        catch( NovaException e ) {
            logger.error(e.getSummary());
            throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("groupId");
        if( blocks.getLength() > 0 ) {
            return blocks.item(0).getFirstChild().getNodeValue().trim();
        }
        throw new CloudException("Failed to create security group without explanation.");
    }
	   
	@Override
	public void delete(@Nonnull String securityGroupId) throws InternalException, CloudException {
		Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.DELETE_SECURITY_GROUP);
		NovaMethod method;
        NodeList blocks;
		Document doc;

        parameters.put("GroupName", securityGroupId);
        /*
        else {
            parameters.put("GroupId", securityGroupId);
        }
        */
		method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
        	doc = method.invoke();
        }
        catch( NovaException e ) {
        	logger.error(e.getSummary());
        	throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("return");
        if( blocks.getLength() > 0 ) {
        	if( !blocks.item(0).getFirstChild().getNodeValue().equalsIgnoreCase("true") ) {
        		throw new CloudException("Failed to delete security group without explanation.");
        	}
        }
	}

	@Override
	public @Nullable
    Firewall getFirewall(@Nonnull String securityGroupId) throws InternalException, CloudException {
        ProviderContext ctx = provider.getContext();
        
        if( ctx == null ) {
            throw new CloudException("No context has been established for this request");
        }
		Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.DESCRIBE_SECURITY_GROUPS);
		NovaMethod method;
        NodeList blocks;
		Document doc;

        parameters.put("GroupName.1", securityGroupId);
        /*
        else {
            parameters.put("GroupId.1", securityGroupId);
        }
        */
		method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
        	doc = method.invoke();
        }
        catch( NovaException e ) {
        	String code = e.getCode();
        	
        	if( code != null && code.startsWith("InvalidGroup") ) {
        		return null;
        	}
        	logger.error(e.getSummary());
        	throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("securityGroupInfo");
        for( int i=0; i<blocks.getLength(); i++ ) {
        	NodeList items = blocks.item(i).getChildNodes();
        	
            for( int j=0; j<items.getLength(); j++ ) {
            	Node item = items.item(j);
            	
            	if( item.getNodeName().equals("item") ) {
            		Firewall firewall = toFirewall(ctx, item);
            		
                    if( firewall != null && securityGroupId.equals(firewall.getProviderFirewallId()) ) {
            			return firewall;
            		}
            	}
            }
        }
        return null;
	}

	@Override
	public @Nonnull String getProviderTermForFirewall(@Nonnull Locale locale) {
		return "security group";
	}

	@Override
	public @Nonnull Collection<FirewallRule> getRules(@Nonnull String securityGroupId) throws InternalException, CloudException {
		Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.DESCRIBE_SECURITY_GROUPS);
		ArrayList<FirewallRule> list = new ArrayList<FirewallRule>();
		NovaMethod method;
        NodeList blocks;
		Document doc;

        parameters.put("GroupName.1", securityGroupId);
        /*
        else {
            parameters.put("GroupId.1", securityGroupId);
        }
        */
		method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
        	doc = method.invoke();
        }
        catch( NovaException e ) {
        	String code = e.getCode();
        	
        	if( code != null && code.startsWith("InvalidGroup") ) {
                return Collections.emptyList();
        	}
        	logger.error(e.getSummary());
        	throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("securityGroupInfo");
        for( int i=0; i<blocks.getLength(); i++ ) {
        	NodeList items = blocks.item(i).getChildNodes();
        	
            for( int j=0; j<items.getLength(); j++ ) {
            	Node item = items.item(j);
            	
            	if( item.getNodeName().equals("item") ) {
            		NodeList attrs = item.getChildNodes();
            		
            		for( int k=0; k<attrs.getLength(); k++ ) {
            			Node attr = attrs.item(k);
            			
            			if( attr.getNodeName().equals("ipPermissions") ) {
            				NodeList subList = attr.getChildNodes();
            				
            				for( int l=0; l<subList.getLength(); l++ ) {
            					Node sub = subList.item(l);
            					
            					if( sub.getNodeName().equals("item") ) {
            						list.addAll(toFirewallRules(securityGroupId, sub));
            					}
            				}
            			}
            		}
            	}
            }
        }
        return list;
	}

	private @Nonnull String getUniqueName(@Nonnull String name) throws InternalException, CloudException {
        StringBuilder str = new StringBuilder();
        
        for( int i=0; i<name.length(); i++ ) {
            char c = name.charAt(i);
            
            if( i == 0 && Character.isDigit(c) ) {
                str.append("e-");
                str.append(c);
            }
            else if( i == 0 && Character.isLetter(c) ) {
                str.append(c);
            }
            else if( i > 0 && (Character.isLetterOrDigit(c) || c == '-' || c == '_') ) {
                str.append(c);
            }
        }
        if( str.length() < 1 ) {
            return "new-group";
        }
        String baseName = str.toString();
        String withName = baseName;
        int count = 1;
        boolean found;
        char c = 'a';
        
        do {
            found = false;
            for( Firewall fw : list() ) {
                String id = fw.getProviderFirewallId();
                
                if( id == null ) {
                    continue;
                }
                if( id.equals(withName) ) {
                    found = true;
                    if( count == 1 ) {
                        withName = baseName + "-" + String.valueOf(c);
                    }
                    else {
                        withName = baseName + String.valueOf(c);
                    }
                    if( c == 'z' ) {
                        if( count == 1 ) {
                            baseName = baseName + "-a";
                        }
                        else {
                            baseName = baseName + "a";
                        }
                        c = 'a';
                        count++;
                        if( count > 10 ) {
                            throw new CloudException("Could not generate a unique firewall name from " + baseName);
                        }
                    }
                    else {
                        c++;
                    }
                    break;
                }                        
            }                        
        } while( found ); 		
        return withName;
	}

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        ComputeServices svc = provider.getComputeServices();
        
        if( svc == null ) {
            return false;
        }
        VirtualMachineSupport support = svc.getVirtualMachineSupport();
        
        return (support != null && support.isSubscribed());
    }

	@Override
	public @Nonnull Collection<Firewall> list() throws InternalException, CloudException {
        ProviderContext ctx = provider.getContext();

        if( ctx == null ) {
            throw new CloudException("No context has been established for this request");
        }
		Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.DESCRIBE_SECURITY_GROUPS);
		ArrayList<Firewall> list = new ArrayList<Firewall>();
		NovaMethod method;
        NodeList blocks;
		Document doc;

		method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
        	doc = method.invoke();
        }
        catch( NovaException e ) {
        	logger.error(e.getSummary());
        	throw new CloudException(e);
        }
        blocks = doc.getElementsByTagName("securityGroupInfo");
        for( int i=0; i<blocks.getLength(); i++ ) {
        	NodeList items = blocks.item(i).getChildNodes();
        	
            for( int j=0; j<items.getLength(); j++ ) {
            	Node item = items.item(j);
            	
            	if( item.getNodeName().equals("item") ) {
            		Firewall firewall = toFirewall(ctx, item);
            		
            		if( firewall != null ) {
            			list.add(firewall);
            		}
            	}
            }
        }
        return list;
	}

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        if( action.equals(FirewallSupport.ANY) ) {
            return new String[] { NovaMethod.EC2_PREFIX + "*" };
        }
        else if( action.equals(FirewallSupport.AUTHORIZE) ) {
            return new String[] { NovaMethod.EC2_PREFIX + NovaMethod.AUTHORIZE_SECURITY_GROUP_INGRESS, NovaMethod.EC2_PREFIX + NovaMethod.AUTHORIZE_SECURITY_GROUP_EGRESS };
        }
        else if( action.equals(FirewallSupport.CREATE_FIREWALL) ) {
            return new String[] { NovaMethod.EC2_PREFIX + NovaMethod.CREATE_SECURITY_GROUP };
        }
        else if( action.equals(FirewallSupport.GET_FIREWALL) || action.equals(FirewallSupport.LIST_FIREWALL) ) {
            return new String[] { NovaMethod.EC2_PREFIX + NovaMethod.DESCRIBE_SECURITY_GROUPS };
        }
        else if( action.equals(FirewallSupport.REMOVE_FIREWALL) ) {
            return new String[] { NovaMethod.EC2_PREFIX + NovaMethod.DELETE_SECURITY_GROUP };
        }
        else if( action.equals(FirewallSupport.REVOKE) ) {
            return new String[] { NovaMethod.EC2_PREFIX + NovaMethod.REVOKE_SECURITY_GROUP_INGRESS, NovaMethod.EC2_PREFIX + NovaMethod.REVOKE_SECURITY_GROUP_EGRESS };
        }
        return new String[0];
    }

	@Override
	public void revoke(@Nonnull String firewallId, @Nonnull String cidr, @Nonnull Protocol protocol, int beginPort, int endPort) throws CloudException, InternalException {
        Map<String,String> parameters = provider.getStandardParameters(provider.getContext(), NovaMethod.REVOKE_SECURITY_GROUP_INGRESS);
        NovaMethod method;
        Document doc;

        parameters.put("GroupName", firewallId);
        /*
        else {
            parameters.put("GroupId", firewallId);
        }
        */
        parameters.put("IpProtocol", protocol.name().toLowerCase());
        parameters.put("FromPort", String.valueOf(beginPort));
        parameters.put("ToPort", endPort == -1 ? String.valueOf(beginPort) : String.valueOf(endPort));
        parameters.put("CidrIp", cidr);
        method = new NovaMethod(provider, provider.getEc2Url(), parameters);
        try {
            doc = method.invoke();
        }
        catch( NovaException e ) {
            logger.error(e.getSummary());
            throw new CloudException(e);
        }
        method.checkSuccess(doc.getElementsByTagName("return"));
    }

    private @Nullable Firewall toFirewall(@Nonnull ProviderContext ctx, @Nullable Node node) {
        if( node == null ) {
            return null;
        }
        String fwName = null, fwId = null, fwDesc = null;
		NodeList attrs = node.getChildNodes();
		Firewall firewall = new Firewall();
        String regionId = ctx.getRegionId();
		String vpcId = null;

        if( regionId == null ) {
            return null;
        }
        firewall.setRegionId(regionId);
        firewall.setAvailable(true);
        firewall.setActive(true);
		for( int i=0; i<attrs.getLength(); i++ ) {
			Node attr = attrs.item(i);
			String name;
			
			name = attr.getNodeName();
			if( name.equals("groupName") ) {
                fwName = attr.getFirstChild().getNodeValue().trim();
			}
			else if( name.equals("groupDescription") ) {
				fwDesc = attr.getFirstChild().getNodeValue().trim();
			}
			else if( name.equals("groupId") ) {
                fwId = attr.getFirstChild().getNodeValue().trim();			    
			}
			else if( name.equals("vpcId") ) {
			    if( attr.hasChildNodes() ) {
			        vpcId = attr.getFirstChild().getNodeValue();
			        if( vpcId != null ) {
			            vpcId = vpcId.trim();
			        }
			    }
			}
		}
        if( fwId == null ) {
            if( fwName == null ) {
                return null;
            }
            fwId = fwName;
        }
		if( fwName == null ) {
		    fwName = fwId;
		}
        firewall.setProviderFirewallId(fwId);
        firewall.setName(fwName);
        if( fwDesc == null ) {
            fwDesc = fwName;
        }
        firewall.setDescription(fwDesc);
		if( vpcId != null ) {
		    firewall.setName(firewall.getName() + " (VPC " + vpcId + ")");
		    firewall.setProviderVlanId(vpcId);
		}
		return firewall;
	}
	
	private @Nonnull Collection<FirewallRule> toFirewallRules(@Nonnull String securityGroupId, @Nullable Node node) {
	    ArrayList<FirewallRule> rules = new ArrayList<FirewallRule>();
        
        if( node == null ) {
            return rules;
        }
	    ArrayList<String> cidrs = new ArrayList<String>();
		NodeList attrs = node.getChildNodes();
        int startPort = -2, endPort = -2;
		Protocol protocol = Protocol.TCP;
		
		for( int i=0; i<attrs.getLength(); i++ ) {
			Node attr = attrs.item(i);
			String name;
			
			name = attr.getNodeName();
			if( name.equals("ipProtocol") ) {
			    String val = attr.getFirstChild().getNodeValue().trim();
			    
			    if( !val.equals("") && !val.equals("-1") ) {
			        protocol = Protocol.valueOf(attr.getFirstChild().getNodeValue().trim().toUpperCase());
			    }
			    else {
			        protocol = Protocol.ICMP;
			    }
			}
			else if( name.equals("fromPort") ) {
				startPort = Integer.parseInt(attr.getFirstChild().getNodeValue().trim());
			}
			else if( name.equals("toPort") ) {
				endPort = Integer.parseInt(attr.getFirstChild().getNodeValue().trim());
			}
			else if( name.equals("ipRanges") ) {
				if( attr.hasChildNodes() ) {
					NodeList children = attr.getChildNodes();
				
					for( int j=0; j<children.getLength(); j++ ) {
						Node child = children.item(j);
					
						if( child.getNodeName().equals("item") ) {
						    if( child.hasChildNodes() ) { 
						        NodeList targets = child.getChildNodes();
						        
						        for( int k=0; k<targets.getLength(); k++ ) {
						            Node cidr = targets.item(k);

						            if( cidr.getNodeName().equals("cidrIp") ) {
						                cidrs.add(cidr.getFirstChild().getNodeValue());
						            }
						        }
						    }
						}
					}
				}
			}
		}
		for( String cidr : cidrs ) {
		    FirewallRule rule = new FirewallRule();

		    rule.setFirewallId(securityGroupId);
		    rule.setCidr(cidr);
		    rule.setEndPort(endPort);
		    rule.setStartPort(startPort);
		    rule.setProtocol(protocol);
		    rules.add(rule);
		}
		return rules;		
	}
}
